import os
import os.path
from tempfile import TemporaryDirectory, mkdtemp
import itertools
import importlib
import logging
import shutil
from typing import List
from contextlib import contextmanager
from pprint import pformat
from time import time

from .transfer import get_downloader, get_uploader
from .archive import unpack_files
from .metadata import extract_product_type_and_level
from .steps import (
    georeference_step, extract_subdataset_step, calc_step, stack_bands_step, output_step
)

logging.basicConfig()

logger = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


def copy_files(source, target):
    for item in os.listdir(source):
        print(item)
        src_path = os.path.join(source, item)
        dst_path = os.path.join(target, item)
        if os.path.isdir(src_path):
            shutil.copytree(
                src_path,
                dst_path
            )
        else:
            shutil.copy(src_path, dst_path)


def custom_preprocessor(source_dir, target_dir, path, args=None, kwargs=None):
    """ Preprocessing step for a custom preprocessing.
    """
    module_name, _, func_name = path.rpartition('.')
    func = getattr(importlib.import_module(module_name), func_name)
    func(source_dir, target_dir, *(args or []), **(kwargs or {}))


def custom_postprocessor(source_dir, target_dir, path, args=None, kwargs=None):
    """ Preprocessing step for a custom preprocessing.
    """
    module_name, _, func_name = path.rpartition('.')
    func = getattr(importlib.import_module(module_name), func_name)
    func(source_dir, target_dir, *(args or []), **(kwargs or {}))


STEP_FUNCTIONS = {
    'custom_preprocessor': custom_preprocessor,
    'subdatasets': extract_subdataset_step,
    'georeference': georeference_step,
    'calc': calc_step,
    'stack_bands': stack_bands_step,
    'output': output_step,
    'custom_postprocessor': custom_postprocessor,
}


def flatten(l):
    return [item for sublist in l for item in sublist]


class Timer:
    """ Helper timer class to allow logging of timing values
    """
    def __init__(self):
        self.start = None
        self.end = None

    def __enter__(self):
        self.start = time()
        return self

    def __exit__(self, *args, **kwargs):
        self.end = time()

    @property
    def elapsed(self):
        return (self.end if self.end is not None else time()) - self.start


# -----------------------------------------------------------------------------

@contextmanager
def workdir(config, use_dir: os.PathLike=None):
    prefix = config.get('prefix', 'preprocess_')
    workdir = config.get('workdir')
    if use_dir:
        os.chdir(use_dir)
        yield use_dir
    elif config.get('keep_temp'):
        dirname = mkdtemp(prefix=prefix, dir=workdir)
        os.chdir(dirname)
        yield dirname
    else:
        with TemporaryDirectory(prefix=prefix, dir=workdir) as dirname:
            os.chdir(dirname)
            yield dirname


def preprocess_file(config: dict, file_path: os.PathLike, use_dir: os.PathLike=None):
    """ Runs the preprocessing of a single file.
    """
    with workdir(config, use_dir) as dirname, Timer() as preprocess_timer:
        logger.info('Preprocessing %s in %s' % (file_path, dirname))

        # check if we can reuse a previous download
        if not os.path.isdir('download'):
            os.mkdir('download')
            logger.info('Downloading %s from %s...' % (file_path, dirname))
            # get the Downloader for the configured source archive to download the given source file
            source_config = config['source']
            downloader = get_downloader(
                source_config['type'], source_config.get('args'), source_config.get('kwargs')
            )

            with Timer() as download_timer:
                source_archive_path = downloader.download(file_path, 'download')

            logger.info(
                'Downloaded file %s in %f seconds'
                % (file_path, download_timer.elapsed)
            )

        else:
            source_archive_path = os.path.join('download', os.path.basename(file_path))
            logger.info('Download dir already exists, skipping...')

        # fetch the metadata XML file from the downloaded archive
        metadata_files = unpack_files(source_archive_path, 'extra', glob=config['metadata_glob'])

        # open the XML to retrieve the product type and level
        product_type, product_level = extract_product_type_and_level(metadata_files, config)
        logger.info('Detected %s/%s' % (product_type, product_level))

        # get a concrete configuration for the type, filled with the defaults
        default_config = dict(config['preprocessing'].get('defaults', {}))
        preprocess_config = dict(config['preprocessing']['types'].get(product_type, {}))
        preprocess_config.update(default_config)

        logger.debug('Using preprocessing config %s' % pformat(preprocess_config))

        if not os.path.isdir('unpack'):
            os.mkdir('unpack')
            logger.info('Unpacking original files...')
            # select and unpack files according to configuration

            with Timer() as unpack_timer:
                data_files = flatten([
                    unpack_files(
                        source_archive_path,
                        'unpack',
                        glob=glob,
                        recursive=preprocess_config.get('nested', False),
                    )
                    for glob in preprocess_config['data_file_globs']
                ])
                metadata_files = flatten([
                    unpack_files(
                        source_archive_path,
                        'unpack',
                        glob=glob,
                        recursive=preprocess_config.get('nested', False),
                    )
                    for glob in preprocess_config.get('additional_file_globs', [])
                ])

            logger.info(
                'Unpacked files: %s in %f seconds'
                % (', '.join(metadata_files + data_files), unpack_timer.elapsed)
            )
        else:
            logger.info('Unpack dir already exists, skipping...')

        previous_step = 'unpack'

        force_refresh = True
        # make processing steps
        for step in ['custom_preprocessor', 'subdatasets', 'georeference', 'calc', 'stack_bands', 'output', 'custom_postprocessor']:
            step_config = preprocess_config.get(step)
            if not step_config:
                logger.debug('Skipping step %s as it is not configured.' % step)
                continue

            # run the step if it was not already run
            if not os.path.isdir(step) or force_refresh:
                if os.path.isdir(step):
                    logger.info('Forcing refresh of existing directory %s' % step)
                    shutil.rmtree(step)

                logger.info('Running preprocessing step %s' % step)
                os.mkdir(step)
                preprpocessor = STEP_FUNCTIONS[step]

                with Timer() as step_timer:
                    preprpocessor(previous_step, step, **step_config)

                logger.info(
                    'Finished preprocessing step %s after %f seconds.'
                    % (step, step_timer.elapsed)
                )

            else:
                logger.info('%s dir already exists, skipping step...' % step)
                force_refresh = True

            previous_step = step

        if not os.path.isdir('upload'):
            os.mkdir('upload')

            # copy files from previous step directory to upload directory
            copy_files(previous_step, 'upload')

        # get an uploader for the finalized images
        target_config = config['target']
        uploader = get_uploader(
            target_config['type'], target_config.get('args'), target_config.get('kwargs')
        )

        upload_filenames = [
            os.path.join(dirpath, filename)
            for dirpath, _, filenames in os.walk('upload')
            for filename in filenames
        ]

        # send all files in the upload directory to the target storage
        logger.info(
            'Starting uploading of %d files to %s'
            % (len(upload_filenames), file_path)
        )
        with Timer() as upload_timer:
            uploader.upload(upload_filenames, file_path)

        logger.info(
            'Finished uploading after %f seconds.'
            % (upload_timer.elapsed)
        )

        logger.info(
            'Finished preprocessing of %s after %f seconds.'
            % (file_path, preprocess_timer.elapsed)
        )

        return upload_filenames
